# --------------------------------------------------------------------------
# Copyright 1992-1993 by Forschungszentrum Informatik (FZI)
#
# You can use and distribute this software under the terms of the license
# version 1 you should have received along with this software.
# If not or if you want additional information, write to
# Forschungszentrum Informatik, "STONE", Haid-und-Neu-Strasse 10-14,
# D-76131 Karlsruhe, Germany.
# --------------------------------------------------------------------------

###################################################################################
# Module:	Maintainance
# Author:	Rainer Neumann
# Date  :	June 1, 1993
# State :	finished
###################################################################################
# Description
# This module provides several maintainance functions for the Object Kernel
###################################################################################

if { [info exists _ok_Maintainance_] == 1 } {
  set load_message "module Maintainance is already present"
} else {
  set _ok_Maintainance_ included
  set load_message "module Maintainance loading"

if { [catch [exec echo {$OK_SRC_PATH}]] } {
  set env(OK_SRC_PATH) [exec pwd]
}

source $env(OK_SRC_PATH)/ObjectKernel.tcl


##########################
# Maintainance functions #
##########################
# The following functions can be used for consistency-checks, etc.

proc okm_getReferencesTo { object } {
  global _ok_ObjectCount
  set hl ""
  for { set i 0 } { $i < $_ok_ObjectCount } { incr i } {
    set al [_ok_getAllAttributes [Object::type "ObjRef_$i"]]
    set x "inst_$i"
    upvar #0 ObjRef_$i $x
    set r ""
    foreach attr $al {
      set ai "($attr)"
      set hs "$x$ai"
      if { [lsearch [set $hs] $object] > -1 } { set r "ObjRef_$i" }
    }
    if { $r != "" } { set hl "$hl $r" }
  }
  return $hl
}

proc okm_showObjectTable { { showRemoved 0 } } {
  global _ok_ObjectCount
  set result "-----------|-------------------------------"
  set result "\n\nObjectID   |  Class\n$result"
  set fm "  "
  for { set i 0 } { $i < $_ok_ObjectCount } { incr i } {
    if { $i == 10 } { set fm " " }
    if { $i == 100 } { set fm "" }
    set x "inst_$i"
    upvar #0 ObjRef_$i $x
    set hs "(Class)"
    set hs "$x$hs"
    set hs [set $hs]
    if { ($hs != "Removed") || $showRemoved } {
      set result "$result\nObjRef_$i$fm |  $hs"
    }
  }
  return "$result\n\n"
}

proc okm_getRemovedObjects { } {
  global _ok_ObjectCount
  set hl ""
  for { set i 0 } { $i < $_ok_ObjectCount } { incr i } {
    set x "inst_$i"
    upvar #0 ObjRef_$i $x
    set hs "(Class)"
    set hs "$x$hs"
    if { [set $hs] == "Removed" } {
      set hl "$hl ObjRef_$i"
    }
  }
  return $hl
}

proc okm_checkIllegalReferences { } {
  set ol [okm_getRemovedObjects]
  set rl ""
  foreach obj $ol {
    set hl [okm_getReferencesTo $obj]
    set rl "$rl [list [list $obj <- $hl]]\n"
  }
  return $rl
}

proc okm_removeObjectsWithIllegalReferences { { recursive 0 } } {
  set rl [okm_checkIllegalReferences]
  foreach ref $rl {
    set ol [lindex $ref 2]
    foreach obj $ol {
      Object::destroy $obj
    }
  }
  if { $recursive && ( $rl != {} ) } {
    okm_removeObjectsWithIllegalReferences 1
  }
}

proc okm_getMetaInfo { classname } {
  global _ok_Classinfo
  set hl $_ok_Classinfo($classname)
  set baseclass [lindex $hl 0]
  if { $baseclass == {} } { set baseclass None }
  return "Classname:  $classname\nBaseclass:  $baseclass\nAttributes:[lindex $hl 1]"
}

proc okm_displayObject { objID { class {} } } {
  if { $class == {} } {
    set metainfo [okm_getMetaInfo [Object::type $objID]]
    set class [lindex $metainfo 1]
    set hs "-----------------------------------------------------"
    set hh "$hs\nShowing Object $objID ($class)\n$hs\n"
  } else {
    set metainfo [okm_getMetaInfo $class]
    set hh ""
  }
  set hs ""
  set attributes [lreplace $metainfo 0 4]
  set ml 0
  foreach attr $attributes {
    set l [string length $attr]
    if { $l > $ml } { set ml $l }
  }
  set result "Class: $class\n"
  if { $class == "Object" } {
    set hs "Self"
    while { [string length $hs] < $ml } { set hs "$hs " }
    set result "$result  $hs    $objID\n"
  }
  foreach attr $attributes {
    set hs $attr
    while { [string length $hs] < $ml } { set hs "$hs " }
    set result "$result  $hs    [$class::get_$attr $objID]\n"
  }
  if { [lindex $metainfo 3] != "None" } {
    set hr [okm_displayObject $objID [lindex $metainfo 3]]
  } else {
    set hr ""
  }
  return "$hh$hr$result"
}


##################
# Memory Cleanup #
##################

proc okm_cleanup_objecttable { {cleanup_interval 0} {periodic 0} } {
  puts stdout "Starting cleanup ..."
  global ObjectCount
  if { $ObjectCount > 0 } {
    set i [expr {$ObjectCount -1}]
    set x "inst_$i"
    upvar #0 ObjRef_$i $x
    set hs "(Class)"
    set hs "$x$hs"
    while { (0 <= $i) && ([set $hs] == "Removed") } {
      incr ObjectCount -1
      incr i -1
      unset $x
      set x "inst_$i"
      upvar #0 ObjRef_$i $x
      set hs "(Class)"
      set hs "$x$hs"
    }
  }
  puts stdout "... done"
  if { ($cleanup_interval > 0) && $periodic } {
    after $cleanup_interval "okm_cleanup_objecttable $cleanup_interval"
  }
}


##########################
# Maintainence Shortcuts #
##########################

proc SOT { {r 0 } } { okm_showObjectTable $r }
proc GRO { } { okm_getRemovedObjects }
proc CIR { } { okm_checkIllegalReferences }
proc ROI { {p {}} } { okm_removeObjectsWithIllegalReferences $p }
proc GRT { o } { okm_getReferencesTo $o }
proc GMI { c } { okm_getMetaInfo $c }
proc DO  { o } { okm_displayObject $o }
proc CM  { o m { p {} } } { _ok_call $o $m $p }
proc COT { {ci 0} {p 0} } { okm_cleanup_objecttable $ci $p }


  set load_message "module Maintainance loaded"
}
